from textwrap import wrap
from ansys.dpf.core.mapping_types import map_types_to_python
from ansys.dpf.core.outputs import _Outputs, Output
from ansys.dpf import core

class Input:
    def __init__(self, spec, pin, operator, count_ellipsis=-1):
        self._spec = spec
        self._operator=operator
        self._pin = pin
        self._count_ellipsis = count_ellipsis
        self._python_expected_types = []
        for cpp_type in self._spec.type_names:
            python_type =map_types_to_python[cpp_type]
            if python_type not in self._python_expected_types:
                self._python_expected_types.append(map_types_to_python[cpp_type])
        if len(self._spec.type_names) == 0:
            self._python_expected_types.append("Any")
        docstr = self.__str__()
        self.name = self._spec.name
        if self._count_ellipsis != -1:
            self.name += str(self._count_ellipsis + 1)
        self._update_doc_str(docstr,self.name)

    def connect(self, inpt):
        """Allows you to connect any input (an entity or an operator
        output) to a specified input pin of this operator.

        Parameters
        ----------
        inpt : str, int, double, Field, FieldsContainer, Scoping, DataSources,
        MeshedRegion, Output, Outputs, Operator
            input of the operator
        """
        # always convert ranges to lists
        if isinstance(inpt, range):
            inpt = list(inpt)
        elif isinstance(inpt, core.Operator):
            if hasattr(inpt, "outputs"):
                inpt = inpt.outputs
            else:
                raise ValueError("The operator to connect in input has no outputs available, it cannot be connected")
        elif isinstance(inpt, core.Model):
            inpt = inpt.metadata.data_sources

        input_type_name = type(inpt).__name__
        if not (input_type_name in self._python_expected_types or ["Outputs", "Output", "Any"]):
            for types in self._python_expected_types:
                print(types, end=' ')
            print("types are expected for", self._spec.name, "pin")
            return

        corresponding_pins = []
        self._operator._find_outputs_corresponding_pins(self._python_expected_types, inpt, self._pin, corresponding_pins)
        if len(corresponding_pins) > 1:
            err_str = "Pin connection is ambiguous, specify the pin with:\n"
            for pin in corresponding_pins:
                err_str += "   - operator.inputs." + self._spec.name+"(out_op." + inpt._dict_outputs[pin[1]].name + ")"
            raise ValueError(err_str)

        if len(corresponding_pins) == 0:
            err_str = f"The input operator for the {self._spec.name} pin be one of the following types:\n"
            err_str += '\n'.join([f'- {py_type}' for py_type in self._python_expected_types])
            raise TypeError(err_str)
            
        if isinstance(inpt, _Outputs):
            self._operator.connect(self._pin, inpt._operator, corresponding_pins[0][1])
            self._operator.inputs._connected_inputs[self._pin] = {corresponding_pins[0][1]: inpt._operator}
        elif isinstance(inpt, Output):
            self._operator.connect(self._pin, inpt._operator, inpt._pin)
            self._operator.inputs._connected_inputs[self._pin] = {inpt._pin: inpt}
        else:
            self._operator.connect(self._pin, inpt)
            self._operator.inputs._connected_inputs[self._pin] = inpt
        
        self.__inc_if_ellipsis()

    def __call__(self, inpt):
        self.connect(inpt)

    def _update_doc_str(self, docstr, class_name):
        """Dynamically updates the docstring of this instance by
        creating a new class on the fly.
        """
        child_class = type(class_name, (self.__class__,), {'__doc__': docstr})
        self.__class__ = child_class

    def __str__(self):
        docstr = self._spec.name + ' : '
        type_info = self._python_expected_types.copy()
        if self._spec.optional:
            type_info += ['optional']
        docstr += ', '.join(type_info) + '\n'
        if self._spec.document:
            docstr += '\n'.join(wrap(self._spec.document.capitalize())) + '\n'
        if self._count_ellipsis >= 0:
            docstr += "is ellipsis\n"
        return docstr

    def __inc_if_ellipsis(self):
        if self._count_ellipsis >=0:
            self._count_ellipsis+=1
            if isinstance(self._operator.inputs, _Inputs):
                self._operator.inputs._add_input(self._pin + self._count_ellipsis, self._spec, self._count_ellipsis)


class _Inputs:
    def __init__(self, dict_inputs, operator):
        self._dict_inputs = dict_inputs
        self._operator = operator
        self._inputs = []
        self._connected_inputs = {}
        
        
    def __str__(self):
        docstr = 'Available inputs:\n'
        for inp in self._inputs:
            tot_string = inp.__str__()
            input_string = tot_string.split('\n')
            input_string1 = input_string[0]
            line = ["   ", "- ", input_string1]
            docstr += '{:<5}{:<4}{:<20}\n'.format(*line)
            for inputstr in input_string:
                if inputstr != input_string1:
                    line = ["   ", "  ", inputstr]
                    docstr += '{:<5}{:<4}{:<20}\n'.format(*line)
        return docstr

    def connect(self, inpt):
        """Allows you to connect any input (an entity or an operator
        output) to any input pin of this operator.

        The matching input type corresponding to the output is looked for.

        Parameters
        ----------
        inpt : str, int, double, Field, FieldsContainer, Scoping, DataSources, MeshedRegion, ScopingsContainer, CyclicSupport, ..., Output, Outputs, Operator
            input of the operator
        """
        corresponding_pins = []
        if isinstance(inpt, core.Operator):
            if hasattr(inpt, "outputs"):
                inpt = inpt.outputs
            else:
                raise ValueError("The operator to connect in input has no outputs available, it cannot be connected")
        elif isinstance(inpt, core.Model):
            inpt = inpt.metadata.data_sources
            
        input_type_name = type(inpt).__name__
        for input_pin in self._inputs:
            self._operator._find_outputs_corresponding_pins(input_pin._python_expected_types, inpt, input_pin._pin, corresponding_pins)
        if len(corresponding_pins) > 1:
            err_str = "Pin connection is ambiguous, specify the pin with:\n"
            for pin in corresponding_pins:
                if isinstance(pin, tuple):
                    pin = pin[0]
                err_str += ("   - operator.inputs." + self._dict_inputs[pin].name+"(input)\n")
            raise ValueError(err_str)

        if len(corresponding_pins) == 0:
            err_str = "The input should have one of the expected types:\n"
            for pin, spec in self._dict_inputs.items():
                for cpp_type in spec.type_names:
                    err_str += f"   - {map_types_to_python[cpp_type]}\n"
            raise TypeError(err_str)

        if not isinstance(inpt,_Outputs) and not isinstance(inpt,Output):
            self._operator.connect(corresponding_pins[0], inpt)
            self._connected_inputs[corresponding_pins[0]] = inpt
        elif input_type_name == "Output":
            self._operator.connect(corresponding_pins[0], inpt._operator, inpt._pin)
            self._connected_inputs[corresponding_pins[0]] = {inpt._pin: inpt._operator}
        else:
            self._operator.connect(corresponding_pins[0][0], inpt._operator, corresponding_pins[0][1])
            self._connected_inputs[corresponding_pins[0][0]]={corresponding_pins[0][1]: inpt._operator}
                

    def _add_input(self, pin, spec, count_ellipsis=-1):
        if spec is not None:
            class_input = Input(spec, pin, self._operator,count_ellipsis)
            class_input.__doc__ = spec.name
            if not hasattr(self, class_input.name):
                setattr(self, class_input.name, class_input)
            else:
                setattr(self, '_' + class_input.name, class_input)
            self._inputs.append(class_input)
            
    def __call__(self, inpt):
        self.connect(inpt)

#Dynamic class Inputs       
class Inputs(_Inputs):
    def __init__(self, dict_inputs, operator):
        super().__init__(dict_inputs, operator)
        
        # dynamically populate input attributes
        for pin, spec in self._dict_inputs.items():
            if spec.ellipsis:
                self._add_input(pin, spec, 0)
            else:
                self._add_input(pin, spec)

    


"""
minus_fc
========
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class minus_fc(Operator):
    """Computes the difference of two fields. If one field's scoping has 'overall' location, then these field's values are applied on the entire other field.

      available inputs:
        - field_or_fields_container_A (FieldsContainer)
        - field_or_fields_container_B (FieldsContainer)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.minus_fc()

      >>> # Make input connections
      >>> my_field_or_fields_container_A = dpf.FieldsContainer()
      >>> op.inputs.field_or_fields_container_A.connect(my_field_or_fields_container_A)
      >>> my_field_or_fields_container_B = dpf.FieldsContainer()
      >>> op.inputs.field_or_fields_container_B.connect(my_field_or_fields_container_B)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.minus_fc(field_or_fields_container_A=my_field_or_fields_container_A,field_or_fields_container_B=my_field_or_fields_container_B)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, field_or_fields_container_A=None, field_or_fields_container_B=None, config=None, server=None):
        super().__init__(name="minus_fc", config = config, server = server)
        self._inputs = InputsMinusFc(self)
        self._outputs = OutputsMinusFc(self)
        if field_or_fields_container_A !=None:
            self.inputs.field_or_fields_container_A.connect(field_or_fields_container_A)
        if field_or_fields_container_B !=None:
            self.inputs.field_or_fields_container_B.connect(field_or_fields_container_B)

    @staticmethod
    def _spec():
        spec = Specification(description="""Computes the difference of two fields. If one field's scoping has 'overall' location, then these field's values are applied on the entire other field.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field_or_fields_container_A", type_names=["fields_container"], optional=False, document="""field or fields container with only one field is expected"""), 
                                 1 : PinSpecification(name = "field_or_fields_container_B", type_names=["fields_container"], optional=False, document="""field or fields container with only one field is expected""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "minus_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMinusFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsMinusFc 
        """
        return super().outputs


#internal name: minus_fc
#scripting name: minus_fc
class InputsMinusFc(_Inputs):
    """Intermediate class used to connect user inputs to minus_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.minus_fc()
      >>> my_field_or_fields_container_A = dpf.FieldsContainer()
      >>> op.inputs.field_or_fields_container_A.connect(my_field_or_fields_container_A)
      >>> my_field_or_fields_container_B = dpf.FieldsContainer()
      >>> op.inputs.field_or_fields_container_B.connect(my_field_or_fields_container_B)
    """
    def __init__(self, op: Operator):
        super().__init__(minus_fc._spec().inputs, op)
        self._field_or_fields_container_A = Input(minus_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field_or_fields_container_A)
        self._field_or_fields_container_B = Input(minus_fc._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._field_or_fields_container_B)

    @property
    def field_or_fields_container_A(self):
        """Allows to connect field_or_fields_container_A input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_field_or_fields_container_A : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus_fc()
        >>> op.inputs.field_or_fields_container_A.connect(my_field_or_fields_container_A)
        >>> #or
        >>> op.inputs.field_or_fields_container_A(my_field_or_fields_container_A)

        """
        return self._field_or_fields_container_A

    @property
    def field_or_fields_container_B(self):
        """Allows to connect field_or_fields_container_B input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_field_or_fields_container_B : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus_fc()
        >>> op.inputs.field_or_fields_container_B.connect(my_field_or_fields_container_B)
        >>> #or
        >>> op.inputs.field_or_fields_container_B(my_field_or_fields_container_B)

        """
        return self._field_or_fields_container_B

class OutputsMinusFc(_Outputs):
    """Intermediate class used to get outputs from minus_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.minus_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(minus_fc._spec().outputs, op)
        self._fields_container = Output(minus_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
Common
===========
"""
from enum import Enum

import numpy as np
import re


from ansys.grpc.dpf import base_pb2
from ansys.grpc.dpf import field_definition_pb2
import progressbar

def _camel_to_snake_case(name):
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


def _remove_spaces(name):
    out = name.lower()
    out = out.replace(" ", "_")
    out = out.replace("-", "_")
    return out


def _snake_to_camel_case(name):
    return ''.join(word.title() for word in name.split('_'))


class _smart_dict_camel(dict):
    def __missing__(self, key):
        return _camel_to_snake_case(key)


class _smart_dict_unit_system(dict):
    def __missing__(self, key):
        return 'unknown'

def __write_enum_doc__(enum, intro=None):
    str =""
    if intro:
        str =intro +" \n\n"
    str+="    Attributes\n"+"    -----------\n"
    for e in enum:
        str+="    "+e.name +" \n\n"
    return str
names = [m.lower() for m in base_pb2.Type.keys()]
names.append("fields_container")
names.append("scopings_container")
names.append("meshes_container")
types = Enum('types', names)
types.__doc__ =__write_enum_doc__(types, "The ``'types'`` enum contains the available types passed through Operators and Workflows to DPF.")

names = [(m.lower(), num)  for m, num in base_pb2.Nature.items()]
natures = Enum('natures', names)
natures.__doc__ =__write_enum_doc__(natures, "The ``'natures'`` enum contains the dimensionnality types.\n It can be used to create a Field of a given dimensionnality.")


names = [(m.lower(), num-1)  for m, num in field_definition_pb2.ShellLayers.items()]
shell_layers = Enum('shell_layers', names)
shell_layers.__doc__ =__write_enum_doc__(shell_layers, "The ``'shell_layers'`` enum contains the available order of shell layers (or lack of shell layers) that defines how the field's data is ordered.")


class locations:
    """Contains locations str for scoping and field locations
    
    Attributes
    -----------
    none = "none"

    elemental = "Elemental"
        data is one per element

    elemental_nodal = "ElementalNodal"
        one per node per element

    nodal = "Nodal"
        one per node

    time_freq = "TimeFreq_sets"
        one per time set
        
    overall = "overall"
        applies everywhere
    
    time_freq_step = "TimeFreq_steps"  
        one per time step      
    """
    none = "none"

    # data is one per element
    elemental = "Elemental"

    # one per node per element
    elemental_nodal = "ElementalNodal"

    # one per node
    nodal = "Nodal"

    # one per time set
    time_freq = "TimeFreq_sets"

    #applies everywhere
    overall = "overall"
    
    # one per time step
    time_freq_step = "TimeFreq_steps"    

    
class DefinitionLabels:
    """Contains Python definition labels."""
    time = "time"
    complex = "complex"
    
def _common_progress_bar(text, unit, tot_size=None):
    
    if tot_size:
        widgets = [progressbar.FormatLabel(f'{text}: %(value)d of %(max_value)d {unit} '),progressbar.Percentage(), progressbar.Bar()]
        return progressbar.ProgressBar(widgets=widgets,max_value=tot_size )
    else:
        widgets = [progressbar.FormatLabel(f'{text}: %(value)d {unit}'), progressbar.RotatingMarker()]
        return progressbar.ProgressBar(widgets=widgets,max_value=progressbar.UnknownLength )



"""
Nodes
=====
"""

import numpy as np


from ansys import dpf
from ansys.dpf.core import  field, property_field
from ansys.grpc.dpf import meshed_region_pb2
from ansys.dpf.core.errors import protect_grpc


class Node:
    """A node of the mesh. Encapsulates all the properties of a node: 
        its id, index, coordinates...

    Created from ``Elements`` or a ``MeshedRegion``.

    Examples
    --------
    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> nodes = model.metadata.meshed_region.nodes

    >>> # Initialize a node from a nodes collection
    >>> node = nodes[0]

    Initialize a node from an element

    >>> element = model.metadata.meshed_region.elements[0]
    >>> node = element.nodes[0]
    
    """

    def __init__(self, mesh, nodeid, index, coordinates):
        self._id = nodeid
        self._index = index
        self._coordinates = coordinates
        self._mesh = mesh

    @property
    def index(self) -> int:
        """Index of the node in the model (int starting at 0)"""
        return self._index

    @property
    def id(self) -> int:
        """Node number"""
        return self._id

    @property
    def coordinates(self):
        """Cartesian coordinates of the node.

        Examples
        --------
        >>> node.coordinates
        [0.015, 0.045, 0.015]
        
        """
        return self._coordinates
    
    @property
    def nodal_connectivity(self):
        """Elements indices connected to the given node

        Returns
        -------
        nodal_connectivity : numpy.array
        """
        return self._mesh.nodes.nodal_connectivity_field.get_entity_data(self.index)

    def __str__(self):
        txt = 'DPF Node     %7d\n' % self.id
        txt += 'Index:      %7d\n' % self.index
        txt += f'Location: {self.coordinates}\n'
        return txt


class Nodes():
    """Collection of DPF Nodes.

    Created from a MeshedRegion

    Examples
    --------
    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> meshed_region = model.metadata.meshed_region
    >>> nodes = model.metadata.meshed_region.nodes
    >>> nodes.n_nodes
    81
    
    """

    def __init__(self, mesh):
        self._mesh = mesh
        self._mapping_id_to_index = None

    def __str__(self):
        return f'DPF Node collection with {len(self)} nodes\n'

    def __getitem__(self, index):
        """Returns node based on index"""
        return self.node_by_index(index)

    def __len__(self):
        return self.n_nodes

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def node_by_id(self, id):
        """Array of node coordinates ordered by index"""
        return self.__get_node(nodeid=id)

    def node_by_index(self, index):
        """Array of node coordinates ordered by index"""
        return self.__get_node(nodeindex=index)

    @protect_grpc
    def __get_node(self, nodeindex=None, nodeid=None):
        """Returns the node by its id or its index

        Parameters
        ----------
        nodeid : int
            id of the requested node

        nodeindex : int
            index of the requested node

        Returns
        -------
        node : ansys.dpf.core.meshed_region.Node
            Requested node
        """
        request = meshed_region_pb2.GetRequest()
        request.mesh.CopyFrom(self._mesh._message)
        if nodeindex is None:
            request.id = nodeid
        else:
            request.index = nodeindex
        nodeOut = self._mesh._stub.GetNode(request)
        return Node(self._mesh, nodeOut.id, nodeOut.index, nodeOut.coordinates)

    @property
    def scoping(self):
        """Return the scoping of the Nodes

        Returns
        -------
        scoping : Scoping
            Scoping of the Nodes

        Examples
        --------
        Get the ids of all the nodes in this collection

        >>> nodes.scoping.ids[2]
        3
        
        """
        return self._mesh._get_scoping(loc=dpf.core.locations.nodal)

    @property
    def n_nodes(self):
        """Number of nodes"""
        return self._mesh._stub.List(self._mesh._message).num_nodes

    @property
    def coordinates_field(self):
        """Coordinates field

        Returns
        -------
        coordinates_field : Field
            field of all the nodes coordinates

        Examples
        --------
        >>> coordinates = nodes.coordinates_field

        >>> # Extract the array of coordinates the coordinates field
        >>> nodes.coordinates_field.data[2]
        array([0.015, 0.045, 0.03 ])
        
        """
        return self._get_coordinates_field()
    
    @property
    def nodal_connectivity_field(self):
        """Nodal connectivity field : field containing for each node id
        the elements indices connected to the given node

        Returns
        -------
        nodal_connectivity_field : PropertyField
            Field of the element indices associated to each node.

        Examples
        --------
        >>> field = nodes.nodal_connectivity_field
        >>> field.get_entity_data(1)
        array([0, 2, 4, 6])
        
        """
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        request.nodal_property = meshed_region_pb2.NODAL_CONNECTIVITY
        fieldOut = self._mesh._stub.ListProperty(request)
        return property_field.PropertyField(server = self._mesh._server, property_field=fieldOut)

    @protect_grpc
    def _get_coordinates_field(self):
        """Return the coordinates field"""
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.nodal_property = meshed_region_pb2.NodalPropertyType.COORDINATES
        request.nodal_property = meshed_region_pb2.COORDINATES
        fieldOut = self._mesh._stub.ListProperty(request)
        return field.Field(server = self._mesh._server, field=fieldOut)

    def _build_mapping_id_to_index(self):
        """Return a mapping between ids and indices of the entity."""
        return {eid: i for i, eid in enumerate(self.scoping.ids)}

    @property
    def mapping_id_to_index(self):
        if self._mapping_id_to_index is None:
            self._mapping_id_to_index = self._build_mapping_id_to_index()
        return self._mapping_id_to_index

    def map_scoping(self, external_scope):
        """Return the indices to map the scoping of these elements to
        the scoping of a field.

        Parameters
        ----------
        external_scope : scoping.Scoping
            Scoping to map to.

        Returns
        -------
        indices : numpy.ndarray
            List of indices to map from the external scope to the
            scoping of these nodes.

        mask : numpy.ndarray
            Members of the external scope that are in the node scoping.

        Examples
        --------
        Return the indices that map a field to a nodes collection.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> nodes = model.metadata.meshed_region.nodes
        >>> disp = model.results.displacement()
        >>> field = disp.outputs.fields_container()[0]
        >>> ind, mask = nodes.map_scoping(field.scoping)

        These indices can then be used to remap ``nodes.coordinates`` to
        match the order of the field data.  That way the field data matches the
        order of the nodes in the ``meshed_region``

        >>> mapped_nodes = nodes.coordinates_field.data[ind]

        """
        if external_scope.location in ['Elemental', 'NodalElemental']:
            raise ValueError('Input scope location must be "Nodal"')
        arr = np.array(list(map(self.mapping_id_to_index.get, external_scope.ids)))
        mask = arr != None
        ind = arr[mask].astype(np.int)
        return ind, mask
    
    def add_node(self, id, coordinates):
        """Appends a new node in the mesh
        
        Parameters
        ----------
        id : int
            new node's id
        
        coordinates : list[float]
            x, y, z coordinates of the node
        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        node_request = meshed_region_pb2.NodeRequest(id=id)
        node_request.coordinates.extend(coordinates)
        request.nodes.append(node_request)
        self._mesh._stub.Add(request)
        
    def add_nodes(self, num):   
        """Add num new nodes in the mesh. 
        add_nodes yields num nodes that the user can fill
        
        Parameters
        ----------
        num : int
            number of nodes to add
        
        Returns
        -------
        yield node : NodeAdder
            node to add
        
        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
        >>> for i, node in enumerate(meshed_region.nodes.add_nodes(4)):
        ...     node.id = i+1
        ...     node.coordinates = [float(i), float(i), 0.0]
        
        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        for i in range(0, num):
            add = NodeAdder()
            yield add
            node_request = meshed_region_pb2.NodeRequest(id=add.id)
            node_request.coordinates.extend(add.coordinates)
            request.nodes.append(node_request)
        self._mesh._stub.Add(request)
        

class NodeAdder:
    """A class used to add new nodes into a meshed region
    
    Attributes
    ----------
    id : int
    
    coordinates : list of doubles
        x, y, z coordinates
    
    Examples
    --------
    Create a meshed region from scratch
    
    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=1)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]       
    ...     i=i+1 
    
    """
    def __init__(self):
        self._id=0
        self._coordinates=[0.0,0.0,0.0]
        
    @property
    def id(self):
        """Node id"""
        return self._id
    
    @id.setter
    def id(self,value):
        self._id =value
        
    @property
    def coordinates(self):
        return self._coordinates
    
    @coordinates.setter
    def coordinates(self, xyz):
        if isinstance(xyz, (np.ndarray, np.generic)):
            self._coordinates= xyz.flatten().tolist()
        else:
            self._coordinates= xyz
            




"""
CyclicSupport
===============
"""
import grpc

from ansys import dpf
from ansys.grpc.dpf import cyclic_support_pb2, cyclic_support_pb2_grpc
from ansys.grpc.dpf import base_pb2, support_pb2
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf import core
from ansys.dpf.core import errors as dpf_errors
from ansys.dpf.core.scoping import Scoping

class CyclicSupport:
    """A class used to represent a CyclicSupport which describes a model with cyclic symmetry.
    It has the necessary data for cyclic (and multistage) expansion
    
    Parameters
    ----------
    cyclic_support : ansys.grpc.dpf.cyclic_support_pb2.CyclicSupport message

    server : DPFServer , optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    Get a CyclicSupport from a model.
    
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> multi_stage = examples.download_multi_stage_cyclic_result()
    >>> model = dpf.Model(multi_stage)
    >>> result_info = model.metadata.result_info
    >>> cyc_support = result_info.cyclic_support
    >>> cyc_support.num_sectors()
    6
    >>> cyc_support.num_stages
    2
    
    """

    def __init__(self, cyclic_support, server=None):
        """Initialize the TimeFreqSupport with its TimeFreqSupport message (if possible)"""
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()
        self._message = cyclic_support

    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)

    @property
    def num_stages(self)->int:
        """Number of cyclic stages in the model

        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> cyc_support.num_stages
        2
        
        """
        return self._stub.List(self._message).num_stages
    
    def num_sectors(self,stage_num=0)->int:
        """Number of sectors to expand on 360°
        
        Parameters
        ----------
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> cyc_support.num_sectors(0)
        6
        >>> cyc_support.num_sectors(1)
        12
        
        """
        return self._stub.List(self._message).stage_infos[stage_num].num_sectors
    
    def base_nodes_scoping(self,stage_num=0)->int:
        """Get a nodal scoping containing nodes ids in the
        base sector of the given stage
        
        Parameters
        ----------
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Returns
        -------
        base_nodes_scoping : Scoping
            nodes ids in the base sector of the given stage
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> base = cyc_support.base_nodes_scoping(0)
            
        """
        return Scoping(scoping=self._stub.List(self._message).stage_infos[stage_num].base_nodes_scoping, server=self._server)
    
    def base_elements_scoping(self,stage_num=0)->int:
        """Get an elemental scoping containing elements ids in the
        base sector of the given stage
        
        Parameters
        ----------
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Returns
        -------
        base_elements_scoping : Scoping
            elements ids in the base sector of the given stage
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> base = cyc_support.base_elements_scoping(stage_num=1)
        
        """
        return Scoping(scoping=self._stub.List(self._message).stage_infos[stage_num].base_elements_scoping, server=self._server)
    
    def sectors_set_for_expansion(self,stage_num=0)->int:
        """Get a sectors scoping (starting from 0 to max = num_sectors-1)
        of the already expanded results and mesh or the the list of sectors that will
        be expanded by default
        
        Parameters
        ----------
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Returns
        -------
        sectors_set_for_expansion : Scoping
            list of sectors (starting from 0 to max = num_sectors-1)
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> print(cyc_support.sectors_set_for_expansion(stage_num=1).ids)
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        
        """
        return Scoping(scoping=self._stub.List(self._message).stage_infos[stage_num].sectors_for_expansion, server=self._server)
    
    
    def expand_node_id(self, node_id, sectors=None, stage_num=0):
        """Returns the node ids corresponding to the base sector node_id given in input
        after expansion
        
        Parameters
        ----------
        node_id : int
            base sector's node id to expand
        
        sectors : Scoping , list of int , optional
            list of sectors to expand (from 0 to max = num_sectors-1)
            default is all the sectors
            
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Returns
        -------
        sectors_set_for_expansion : Scoping
            list of sectors (starting from 0 to max = num_sectors-1)
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> print(cyc_support.expand_node_id(1,stage_num=0).ids)
        [1, 3596, 5816, 8036, 10256, 12476]
        
        """
        if isinstance(sectors, list):
            sectors = Scoping(ids=sectors, location="sectors",server=self._server)
        
        request = cyclic_support_pb2.GetExpandedIdsRequest()
        request.support.CopyFrom(self._message)
        request.node_id = node_id
        request.stage_num =stage_num
        if sectors:
            request.sectors_to_expand.CopyFrom(sectors._message)
        return Scoping(scoping=self._stub.GetExpandedIds(request).expanded_ids, server=self._server)
    
    def expand_element_id(self, element_id, sectors=None, stage_num=0):
        """Returns the element ids corresponding to the base sector element_id given in input
        after expansion
        
        Parameters
        ----------
        element_id : int
            base sector's element id to expand
        
        sectors (optional) : Scoping or list of int
            list of sectors to expand (from 0 to max = num_sectors-1)
            default is all the sectors
            
        stage_num : int , optional
            number of the stage required (from 0 to num_stages)
            
        Returns
        -------
        sectors_set_for_expansion : Scoping
            list of sectors (starting from 0 to max = num_sectors-1)
            
        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> multi_stage = examples.download_multi_stage_cyclic_result()
        >>> cyc_support = Model(multi_stage).metadata.result_info.cyclic_support
        >>> print(cyc_support.expand_element_id(1,stage_num=0).ids)
        [1, 1558, 2533, 3508, 4483, 5458]
        
        """
        if isinstance(sectors, list):
            sectors = Scoping(ids=sectors, location="sectors",server=self._server)
        
        request = cyclic_support_pb2.GetExpandedIdsRequest()
        request.support.CopyFrom(self._message)
        request.element_id = element_id
        request.stage_num =stage_num
        if sectors:
            request.sectors_to_expand.CopyFrom(sectors._message)
        return Scoping(scoping=self._stub.GetExpandedIds(request).expanded_ids, server=self._server)
    
    def _connect(self):
        """Connect to the grpc service"""
        return cyclic_support_pb2_grpc.CyclicSupportServiceStub(self._server.channel)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass



import sys
import inspect

## to do : change that one the module is done
from ansys.dpf.core.meshed_region import *
from ansys.dpf.core.available_result import *
from ansys.dpf.core.data_sources import *
from ansys.dpf.core.field import *
from ansys.dpf.core.fields_container import *
from ansys.dpf.core.scoping import *
from ansys.dpf.core.time_freq_support import *
from ansys.dpf.core.common import (_smart_dict_camel,
                                   _camel_to_snake_case,
                                   _snake_to_camel_case,
                                   _smart_dict_unit_system)


map_types_to_cpp = _smart_dict_camel()
for classes in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    map_types_to_cpp[classes[0]] = _camel_to_snake_case(classes[0])
map_types_to_cpp['str'] = 'string'
map_types_to_cpp['MeshedRegion'] = 'abstract_meshed_region'
map_types_to_cpp['list'] = 'vector<int32>'
map_types_to_cpp['bool'] = 'bool'
map_types_to_cpp['int'] = 'int32'
map_types_to_cpp['double'] = 'double'
map_types_to_cpp['float'] = 'double'


class _smart_dict_snake(dict):
    def __missing__(self, key):
        return _snake_to_camel_case(key)
map_types_to_python = _smart_dict_snake()
for k, v in map_types_to_cpp.items():
    map_types_to_python[v] = k
map_types_to_python['vector<double>']= 'list'

map_unit_system = _smart_dict_unit_system()
map_unit_system[5] = 'Metric (cm, g, dyne, s, V, A)'
map_unit_system[6] = 'Metric (mm, Kg, N, s, mV, mA)'
map_unit_system[7] = 'U.S Customary (ft, lbm, lbf, F, s, V, A)'
map_unit_system[8] = 'U.S Customary (in, lbm, lbf, F, s, V, A)'
map_unit_system[10] = 'Metric (um, t, uN, s, V, mA)'
map_unit_system[11] = 'Metric (m, kg, N, s, V, A)'



"""
nodal_difference
================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "averaging" category
"""

class nodal_difference(Operator):
    """Transform ElementalNodal field into Nodal field. Each nodal value is the maximum difference between the unaveraged computed result for all elements that share this particular node. Result is computed on a given node scoping.

      available inputs:
        - field (Field, FieldsContainer)
        - mesh_scoping (Scoping) (optional)
        - mesh (MeshedRegion) (optional)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.averaging.nodal_difference()

      >>> # Make input connections
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.averaging.nodal_difference(field=my_field,mesh_scoping=my_mesh_scoping,mesh=my_mesh)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, field=None, mesh_scoping=None, mesh=None, config=None, server=None):
        super().__init__(name="nodal_difference", config = config, server = server)
        self._inputs = InputsNodalDifference(self)
        self._outputs = OutputsNodalDifference(self)
        if field !=None:
            self.inputs.field.connect(field)
        if mesh_scoping !=None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Transform ElementalNodal field into Nodal field. Each nodal value is the maximum difference between the unaveraged computed result for all elements that share this particular node. Result is computed on a given node scoping.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field","fields_container"], optional=False, document="""field or fields container with only one field is expected"""), 
                                 1 : PinSpecification(name = "mesh_scoping", type_names=["scoping"], optional=True, document="""average only on these entities"""), 
                                 7 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=True, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "nodal_difference")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsNodalDifference 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsNodalDifference 
        """
        return super().outputs


#internal name: nodal_difference
#scripting name: nodal_difference
class InputsNodalDifference(_Inputs):
    """Intermediate class used to connect user inputs to nodal_difference operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.averaging.nodal_difference()
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(nodal_difference._spec().inputs, op)
        self._field = Input(nodal_difference._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field)
        self._mesh_scoping = Input(nodal_difference._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._mesh_scoping)
        self._mesh = Input(nodal_difference._spec().input_pin(7), 7, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def field(self):
        """Allows to connect field input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_field : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.averaging.nodal_difference()
        >>> op.inputs.field.connect(my_field)
        >>> #or
        >>> op.inputs.field(my_field)

        """
        return self._field

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator

        - pindoc: average only on these entities

        Parameters
        ----------
        my_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.averaging.nodal_difference()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> #or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)

        """
        return self._mesh_scoping

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.averaging.nodal_difference()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsNodalDifference(_Outputs):
    """Intermediate class used to get outputs from nodal_difference operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.averaging.nodal_difference()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(nodal_difference._spec().outputs, op)
        self._fields_container = Output(nodal_difference._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.averaging.nodal_difference()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
Check the matching for a client/server pair. 

Used to verify if the server version is a minimum value. 
"""

from ansys.dpf.core import errors as dpf_errors
import sys

def server_meet_version(required_version, server):
    """
    Check if a given server version matches with a required version.
    
    Parameters
    ----------
    required_version : str
        Required version that will be compared with the server version.
        
    server : Server
        Dpf server object.

    Returns
    -------
    bool : 
        True if the server version meets the requirement.
    """
    version = get_server_version(server)
    meets = version_tuple(required_version)
    return meets_version(version, meets)


def server_meet_version_and_raise(required_version, server, msg = None):
    """
    Check if a given server version matches with a required version and throws 
    if it doesn't.
    
    Parameters
    ----------
    required_version : str
        Required version that will be compared with the server version.
    server : Server
        Dpf server object.
    msg : str, optional
        Message to be contained in the raised Exception if versions are
        not meeting.

    Raises
    ------
    dpf_errors : errors
        errors.DpfVersionNotSupported is raised if failure.

    Returns
    -------
    bool : 
        True if the server version meets the requirement.
    """
    
    if not server_meet_version(required_version,server):
        if msg is not None:
            raise dpf_errors.DpfVersionNotSupported(required_version, msg=msg)
        else: 
            raise dpf_errors.DpfVersionNotSupported(required_version)
    # server meets version if ends here
    return True
            
            
def meets_version(version, meets):
    """
    Check if a version string meets a minimum version.

    Parameters
    ----------
    version : str
        Version to check.
        For example ``'1.32.1'``.
        
    meets : str
        Required version (version must be compared to it).
        For example ``'1.32.2'``.

    Returns
    -------
    bool :
        True if the server version meets the requirements.
    """
    if not isinstance(version, tuple):
        va = version_tuple(version)
    else:
        va = version
        
    if not isinstance(meets, tuple):
        vb = version_tuple(meets)
    else:
        vb = meets

    if len(va) != len(vb):
        raise ValueError("Versions are not comparable.")

    for i in range(len(va)):
        if va[i] > vb[i]:
            return True
        elif va[i] < vb[i]:
            return False

    return True
            
def get_server_version(server = None):
    """Return server version as a string."""
    from ansys.dpf import core
    if server is None:
        version = core.SERVER.version
    else:
        version = server.version
    return version


def version_tuple(ver):
    """Convert a version string to a tuple containing ints.
    
    Parameters
    ----------
    ver : str
    
    Returns
    -------
    ver_tuple : tuple
        Length 3 tuple representing the major, minor, and patch
        version.
    """
    split_ver = ver.split(".")
    while len(split_ver) < 3:
        split_ver.append('0')

    if len(split_ver) > 3:
        raise ValueError('Version strings containing more than three parts '
                         'cannot be parsed')

    vals = []
    for item in split_ver:
        if item.isnumeric():
            vals.append(int(item))
        else:
            vals.append(0)

    return tuple(vals)

            
def version_requires(min_version):
    """Must be used as decorator. 
    Ensure the method called matches a certain server version."""

    def decorator(func):
        # first arg *must* be a tuple containing the version
        if not isinstance(min_version, str):
            raise TypeError('version_requires decorator must a string with dot separator.')

        def wrapper(self, *args, **kwargs):
            """Call the original function"""
            server = self._server
            func_name = func.__name__
            class_name = self.__class__.__name__
            
            # particular cases
            # scoping._set_ids case, must be checked in a particular way
            if func_name == "_set_ids" and class_name == "Scoping":
                ids = args[0]
                size = len(ids)
                if size != 0:
                    max_size = 8.0E6//sys.getsizeof(ids[0])
                    if size > max_size:
                        server.check_version(min_version)
            # default case, just check the compatibility
            else: 
                server.check_version(min_version)

            return func(self, *args, **kwargs)

        return wrapper

    return decorator
    


import logging


def setup_logger(loglevel='INFO'):
    """ Setup logger """

    # return existing log if this function has already been called
    if hasattr(setup_logger, 'log'):
        setup_logger.log.setLevel(loglevel)
        ch = setup_logger.log.handlers[0]
        ch.setLevel(loglevel)
        return setup_logger.log

    # create logger
    log = logging.getLogger(__name__)
    log.setLevel(loglevel)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(loglevel)

    # create formatter
    formatstr = '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
    formatter = logging.Formatter(formatstr)

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    log.addHandler(ch)

    # make persistent
    setup_logger.log = log

    return log



"""
.. _ref_fields_container:
    
FieldsContainer
===============
Contains classes associated to the DPF FieldsContainer
"""
from ansys import dpf
from ansys.dpf.core.collection import Collection
from ansys.dpf.core.common import types
from ansys.dpf.core import errors as dpf_errors


class FieldsContainer(Collection):
    """A class used to represent a FieldsContainer which contains
    fields belonging to a common results.
    The fields container is designed as a set of fields ordered by labels 
    and ids. Each field of the Fields Container has an id for each label 
    defining the given Fields Container. This allows to split the fields 
    on any criteria.
    The most common fields container have the label "time" with ids 
    corresponding to each time sets, the label "complex" will allow 
    to separate real parts (id=0) from imaginary parts (id=1) 
    in a harmonic analysis for example. 

    Parameters
    ----------
    fields_container : ansys.grpc.dpf.collection_pb2.Collection or FieldsContainer, optional
        Create a fields container from a Collection message or create a copy from an existing fields container

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
    
    Examples
    --------
    Extract a displacement fields container from a transient result file.

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = dpf.Model(transient)
    >>> disp = model.results.displacement()
    >>> disp.inputs.time_scoping.connect([1,5])
    >>> fields_container = disp.outputs.fields_container()
    >>> field_set_5 =fields_container.get_fields_by_time_complex_ids(5)
    >>> #print(fields_container)

    
    Create a fields container from scratch
    
    >>> from ansys.dpf import core as dpf
    >>> fc= dpf.FieldsContainer()
    >>> fc.labels =['time','complex']
    >>> for i in range(0,20): #real fields 
    ...     mscop = {"time":i+1,"complex":0}
    ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
    >>> for i in range(0,20): #imaginary fields
    ...     mscop = {"time":i+1,"complex":1}
    ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
    
    """

    def __init__(self, fields_container=None, server=None):
        """Initialize the scoping with either optional scoping message,
        or by connecting to a stub.
        """
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        self._component_index = None  # component index
        self._component_info = None  # for norm/max/min

        Collection.__init__(self, types.field,
                            collection=fields_container, server=self._server)

    def get_fields_by_time_complex_ids(self, timeid=None, complexid=None):
        """Returns the fields at a requested time/freq id and real or imaginary depending on 
        complexid (complexid=1:imaginary, complexid=0:real)

        Parameters
        ----------
        timeid : int, optional
            The time id. One based index of the result set.

        complexid : int, optional
            The complex id.

        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
            
        Examples
        --------
        Extract a the 5th time set of a transient analysis.
    
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> len(model.metadata.time_freq_support.time_frequencies)
        35
        >>> disp = model.results.displacement()
        >>> disp.inputs.time_scoping.connect([1,5])
        >>> fields_container = disp.outputs.fields_container()
        >>> field_set_5 =fields_container.get_fields_by_time_complex_ids(5)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        return super()._get_entries(label_space)
    
    def get_field_by_time_complex_ids(self, timeid=None, complexid=None):
        """Returns the field at a requested time/freq id and real or imaginary depending on 
        complexid (complexid=1:imaginary, complexid=0:real).
        It throws if the number of fields matching the request is 
        higher than 1.

        Parameters
        ----------
        timeid : int, optional
            The time id. One based index of the result set.

        complexid : int, optional
            The complex id.

        Returns
        -------
        fields : Field
            field corresponding to the request
            
        Examples
        --------
        Extract a the 5th time set of a transient analysis.
    
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> len(model.metadata.time_freq_support.time_frequencies)
        35
        >>> disp = model.results.displacement()
        >>> disp.inputs.time_scoping.connect([1,5])
        >>> fields_container = disp.outputs.fields_container()
        >>> field_set_5 =fields_container.get_fields_by_time_complex_ids(5)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        return super()._get_entry(label_space)
    
    def __time_complex_label_space__(self, timeid=None, complexid=None):
        label_space ={}
        if timeid is not None:
            label_space["time"] = timeid
        if complexid is not None:
            label_space["complex"] = complexid
        return label_space

    def get_fields(self, label_space):
        """Returns the fields at a requested index or label space

        Parameters
        ----------
        label_space : dict[str,int] 
            Scoping of the requested fields, for example:
            ``{"time": 1, "complex": 0}``

        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
          
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> fc= dpf.FieldsContainer()
        >>> fc.labels =['time','complex']
        >>> #real fields 
        >>> for i in range(0,20): 
        ...     mscop = {"time":i+1,"complex":0}
        ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
        >>> #imaginary fields
        >>> for i in range(0,20):
        ...     mscop = {"time":i+1,"complex":1}
        ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
                
        >>> fields = fc.get_fields({"time":2})
        >>> # imaginary and real fields of time 2
        >>> len(fields)
        2
        
        """
        
        return super()._get_entries(label_space)
    
    def get_field(self, label_space_or_index):
        """Returns the field at a requested index or label space.
        Throws if the request returns several fields

        Parameters
        ----------
        label_space_or_index : dict[str,int] , int 
            Scoping of the requested fields, for example:
            ``{"time": 1, "complex": 0}``
            or Index of the field.

        Returns
        -------
        field : Field
            field corresponding to the request
          
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> fc= dpf.fields_container_factory.over_time_freq_fields_container([dpf.Field(nentities=10)])
        >>> field = fc.get_field({"time":1})

        """        
        return super()._get_entry(label_space_or_index)
    
    
    def get_field_by_time_id(self, timeid=None):
        """Returns the complex field at a requested time

        Parameters
        ----------
        timeid: int, optional
            The time id. Index of the result set.

        Returns
        -------
        fields : Field 
            fields corresponding to the request
        """
        if (not self.has_label('time')):
            raise dpf_errors.DpfValueError("The fields container is not based on time scoping.")
        
        if self.has_label('complex'):
             label_space = self.__time_complex_label_space__(timeid,0)
        else:
            label_space = self.__time_complex_label_space__(timeid)
        
        return super()._get_entry(label_space)
    
    def get_imaginary_fields(self, timeid=None):
        """Returns the complex fields at a requested time

        Parameters
        ----------
        timeid: int, optional
            The time id. Index of the result set.

        Returns
        -------
        fields : list[Field] 
            fields corresponding to the request
        """
        if (not self.has_label('complex') or not self.has_label('time')):
            raise dpf_errors.DpfValueError("The fields container is not based on time and complex scoping.")
        
        label_space = self.__time_complex_label_space__(timeid,1)
        
        return super()._get_entries(label_space)
    
    def get_imaginary_field(self, timeid=None):
        """Returns the complex field at a requested time

        Parameters
        ----------
        timeid: int, optional
            The time id. Index of the result set.

        Returns
        -------
        fields : Field 
            field corresponding to the request
        """
        if (not self.has_label('complex') or not self.has_label('time')):
            raise dpf_errors.DpfValueError("The fields container is not based on time and complex scoping.")
        
        label_space = self.__time_complex_label_space__(timeid,1)
        
        return super()._get_entry(label_space)

    def __getitem__(self, key):
        """Returns the field at a requested index

        Parameters
        ----------
        key : int
            the index

        Returns
        -------
        field : Field
            field corresponding to the request
        """
        return super().__getitem__(key)

    def add_field(self, label_space, field):
        """Update or add the field at a requested label space.

        Parameters
        ----------
        label_space : dict[str,int]
            label_space of the requested fields, ex : {"time":1, "complex":0}

        field : Field
            DPF field to add.
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> fc= dpf.FieldsContainer()
        >>> fc.labels =['time','complex']
        >>> for i in range(0,20): #real fields 
        ...     mscop = {"time":i+1,"complex":0}
        ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
        >>> for i in range(0,20): #imaginary fields
        ...     mscop = {"time":i+1,"complex":1}
        ...     fc.add_field(mscop,dpf.Field(nentities=i+10))
                
        """
        super()._add_entry(label_space, field)

    def add_field_by_time_id(self, field, timeid = 1):
        """Update or add the field at a requested time id.

        Parameters
        ----------        
        field : Field
            DPF field to add.
            
        timeid: int, optional
            time id corresponding to the requested time set. Default: 1
        """
        labels = self.labels
        if (not self.has_label('time') 
            and (len(self.labels) == 0 
            or (len(self.labels) == 1 and self.has_label("complex")))):
            self.add_label('time')
        if (len(self.labels) == 1):
            super()._add_entry({'time': timeid}, field)
        elif (self.has_label('time') and self.has_label('complex') and len(labels) == 2):
            super()._add_entry({'time': timeid, 'complex': 0}, field)
        else:
            raise dpf_errors.DpfValueError('The fields container is not only based on time scoping.')
            
    def add_imaginary_field(self, field, timeid = 1):
        """Update or add the imaginary field at a requested time id.

        Parameters
        ----------        
        field : Field
            DPF field to add.
            
        timeid: int, optional
            time id corresponding to the requested time set. Default: 1
        """
        if (not self.has_label('time') 
            and (len(self.labels) == 0 
            or (len(self.labels) == 1 and self.has_label("complex")))):
            self.add_label('time')
        if (not self.has_label('complex') and len(self.labels) == 1 and self.has_label('time')):
            self.add_label('complex')
        if (self.has_label('time') and self.has_label('complex') and len(self.labels) == 2):
            super()._add_entry({'time': timeid, 'complex': 1}, field)
        else:
            raise dpf_errors.DpfValueError('The fields container is not only based on time scoping.')

    def select_component(self, index):
        """Returns fields containing only the component index.

        Can only select by component index as multiple fields may
        contain different number of components.

        Parameters
        ----------
        index : int
            Component index.

        Returns
        -------
        fields : FieldsContainer
            Fields container with one component selected in each field.

        Examples
        --------
        Select using a component index
    
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> disp.inputs.time_scoping.connect([1,5])
        >>> fields_container = disp.outputs.fields_container()
        >>> disp_x_fields = fields_container.select_component(0)
        >>> my_field = disp_x_fields[0]
          
        """
        comp_select = dpf.core.Operator("component_selector_fc")
        comp_select.connect(0,self)
        comp_select.connect(1,index)
        return comp_select.outputs.fields_container.get_data()

    @property
    def time_freq_support(self):
        return self._get_time_freq_support()
    
    @time_freq_support.setter
    def time_freq_support(self, value):
        return super()._set_time_freq_support(value)
    
    
    def deep_copy(self,server=None):
        """Creates a deep copy of the fields container's data (and its fields) on a given server.
        This can be useful to pass data from one server instance to another.
        
        Parameters
        ----------
        server : DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the the global server.
        
        Returns
        -------
        fields_container_copy : FieldsContainer
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> disp.inputs.time_scoping.connect([1,5])
        >>> fields_container = disp.outputs.fields_container()
        >>> other_server = dpf.start_local_server(as_global=False)
        >>> deep_copy = fields_container.deep_copy(server=other_server)
        
        """
        fc = FieldsContainer(server=server)
        fc.labels= self.labels
        for i,f in enumerate(self):
            fc.add_field(self.get_label_space(i),f.deep_copy(server))        
        try:
            fc.time_freq_support = self.time_freq_support.deep_copy(server)
        except:
            pass
        return fc
    
    def get_time_scoping(self):
        """Returns the time scoping containing the time sets
        
        Returns
        -------
        scoping: Scoping
            scoping containing the time set ids available in the fields container
        """
        return self.get_label_scoping("time")
    
    def __add__(self, fields_b):
        """Adds two fields or fields containers together
                
        Returns
        -------
        add : operators.math.add_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "add_fc") :
            op= operators.math.add_fc(self, fields_b)
        else :
            op= dpf_operator.Operator("add_fc")
            op.connect(0,self)        
            op.connect(1, fields_b)
        return op
    
    def __sub__(self, fields_b):
        """Subtract two fields or fields containers together
                
        Returns
        -------
        minus : operators.math.minus_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "minus_fc") :
            op= operators.math.minus_fc()
        else :
            op= dpf_operator.Operator("minus_fc")
        op.connect(0,self)        
        op.connect(1, fields_b)
        return op

    def __pow__(self, value):
        if value != 2:
            raise ValueError('DPF only the value is "2" suppported')
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "sqr_fc") :
            op= operators.math.sqr_fc()
        else :
            op= dpf_operator.Operator("sqr_fc")
        op.connect(0,self)        
        op.connect(1, value)
        return op
    
    def __mul__(self, value):
        """Multiplies two fields or fields containers together
        
        Returns
        -------
        mul : operators.math.generalized_inner_product_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "generalized_inner_product_fc") :
            op= operators.math.generalized_inner_product_fc()
        else :
            op= dpf_operator.Operator("generalized_inner_product_fc")
        op.connect(0,self)        
        op.connect(1, value)
        return op
    
    



"""Installation file for python dpf module
"""
import os
from io import open as io_open

from setuptools import setup

install_requires = ['psutil',
                    'progressbar2',
                    'numpy',
                    'ansys.grpc.dpf>=0.2.3']


# Get version from version info
filepath = os.path.dirname(__file__)
__version__ = None
version_file = os.path.join(filepath, 'ansys', 'dpf', 'core', '_version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())  # execute file from raw string


readme_file = os.path.join(filepath, 'README.md')

setup(
    name='ansys-dpf-core',
    packages=['ansys.dpf.core', 'ansys.dpf.core.examples', 'ansys.dpf.core.operators', 
    'ansys.dpf.core.operators.averaging', 'ansys.dpf.core.operators.filter', 'ansys.dpf.core.operators.geo', 
    'ansys.dpf.core.operators.invariant', 'ansys.dpf.core.operators.logic', 'ansys.dpf.core.operators.mapping', 
    'ansys.dpf.core.operators.math', 'ansys.dpf.core.operators.mesh', 'ansys.dpf.core.operators.metadata', 
    'ansys.dpf.core.operators.min_max', 'ansys.dpf.core.operators.result', 'ansys.dpf.core.operators.scoping', 
    'ansys.dpf.core.operators.serialization', 'ansys.dpf.core.operators.utility'],
    version=__version__,
    description='DPF Python gRPC client',
    # long_description=io_open(readme_file, encoding="utf-8").read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    package_data={'ansys.dpf.core.examples': ['ASimpleBar.rst',
                                              'static.rst',
                                              'complex.rst',
                                              'model_with_ns.rst',
                                              'file_cyclic.rst',
                                              'msup_transient_plate1.rst',
                                              'rth/rth_electric.rth',
                                              'rth/rth_steady.rth',
                                              'rth/rth_transient.rth',
                                              'sub/cp56.sub', 
                                              'msup/file.mode',
                                              'msup/file.rst',
                                              'msup/file.rfrq',
                                              'distributed/file0.rst',
                                              'distributed/file1.rst',
    ]},

    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
    install_requires=install_requires,
    extras_require={
        "plotting":  ['pyvista>=0.24.0', 'matplotlib==3.2'],
        "reporting":  ['scooby'],
    }
)



"""
principal_invariants_fc
=======================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "invariant" category
"""

class principal_invariants_fc(Operator):
    """Computes the element-wise eigen values of all the tensor fields of a fields container.

      available inputs:
        - fields_container (FieldsContainer)

      available outputs:
        - fields_eig_1 (FieldsContainer)
        - fields_eig_2 (FieldsContainer)
        - fields_eig_3 (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.invariant.principal_invariants_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.invariant.principal_invariants_fc(fields_container=my_fields_container)

      >>> # Get output data
      >>> result_fields_eig_1 = op.outputs.fields_eig_1()
      >>> result_fields_eig_2 = op.outputs.fields_eig_2()
      >>> result_fields_eig_3 = op.outputs.fields_eig_3()"""
    def __init__(self, fields_container=None, config=None, server=None):
        super().__init__(name="invariants_fc", config = config, server = server)
        self._inputs = InputsPrincipalInvariantsFc(self)
        self._outputs = OutputsPrincipalInvariantsFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)

    @staticmethod
    def _spec():
        spec = Specification(description="""Computes the element-wise eigen values of all the tensor fields of a fields container.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_eig_1", type_names=["fields_container"], optional=False, document="""first eigen value fields"""), 
                                 1 : PinSpecification(name = "fields_eig_2", type_names=["fields_container"], optional=False, document="""second eigen value fields"""), 
                                 2 : PinSpecification(name = "fields_eig_3", type_names=["fields_container"], optional=False, document="""third eigen value fields""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "invariants_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPrincipalInvariantsFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsPrincipalInvariantsFc 
        """
        return super().outputs


#internal name: invariants_fc
#scripting name: principal_invariants_fc
class InputsPrincipalInvariantsFc(_Inputs):
    """Intermediate class used to connect user inputs to principal_invariants_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.invariant.principal_invariants_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
    """
    def __init__(self, op: Operator):
        super().__init__(principal_invariants_fc._spec().inputs, op)
        self._fields_container = Input(principal_invariants_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.principal_invariants_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

class OutputsPrincipalInvariantsFc(_Outputs):
    """Intermediate class used to get outputs from principal_invariants_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.invariant.principal_invariants_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_eig_1 = op.outputs.fields_eig_1()
      >>> result_fields_eig_2 = op.outputs.fields_eig_2()
      >>> result_fields_eig_3 = op.outputs.fields_eig_3()
    """
    def __init__(self, op: Operator):
        super().__init__(principal_invariants_fc._spec().outputs, op)
        self._fields_eig_1 = Output(principal_invariants_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_eig_1)
        self._fields_eig_2 = Output(principal_invariants_fc._spec().output_pin(1), 1, op) 
        self._outputs.append(self._fields_eig_2)
        self._fields_eig_3 = Output(principal_invariants_fc._spec().output_pin(2), 2, op) 
        self._outputs.append(self._fields_eig_3)

    @property
    def fields_eig_1(self):
        """Allows to get fields_eig_1 output of the operator


        - pindoc: first eigen value fields

        Returns
        ----------
        my_fields_eig_1 : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.principal_invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_eig_1 = op.outputs.fields_eig_1() 
        """
        return self._fields_eig_1

    @property
    def fields_eig_2(self):
        """Allows to get fields_eig_2 output of the operator


        - pindoc: second eigen value fields

        Returns
        ----------
        my_fields_eig_2 : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.principal_invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_eig_2 = op.outputs.fields_eig_2() 
        """
        return self._fields_eig_2

    @property
    def fields_eig_3(self):
        """Allows to get fields_eig_3 output of the operator


        - pindoc: third eigen value fields

        Returns
        ----------
        my_fields_eig_3 : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.principal_invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_eig_3 = op.outputs.fields_eig_3() 
        """
        return self._fields_eig_3




from ansys.dpf.core.mapping_types import map_types_to_python
from ansys.dpf.core.common import types
from ansys.grpc.dpf import operator_pb2
import re

class Output:
    def __init__(self, spec, pin, operator):
        self._spec = spec
        self._operator = operator
        self._pin = pin
        self._python_expected_types = []
        for cpp_type in self._spec.type_names:
            self._python_expected_types.append(map_types_to_python[cpp_type])

    def get_data(self):
        """Returns this output of this operator"""
        type_output = self._spec.type_names[0]
        if type_output == 'abstract_meshed_region':
            type_output = 'meshed_region'

        elif type_output == "fields_container":
            type_output = types.fields_container
        elif type_output == "scopings_container":
            type_output = types.scopings_container
        elif type_output == "meshes_container":
            type_output = types.meshes_container

        return self._operator.get_output(self._pin, type_output)

    def __call__(self):
        return self.get_data()

    def __str__(self):
        docstr = self._spec.name
        if self._spec.optional:
            docstr += " (optional)"
        docstr += ", expects types:"+'\n'
        for exp_types in self._python_expected_types:
            docstr += "   -" + exp_types + '\n'
        if self._spec.document:
            docstr += "help: " + self._spec.document + '\n'
        return docstr


class _Outputs:
    def __init__(self, dict_outputs, operator):
        self._dict_outputs = dict_outputs
        self._operator = operator
        self._outputs = []
        
    def _get_given_output(self, input_type_name):
        corresponding_pins = []
        for asked_camel_types in input_type_name:
            for pin in self._dict_outputs:
                for cpp_type in self._dict_outputs[pin].type_names:
                    if map_types_to_python[cpp_type] == asked_camel_types:
                        corresponding_pins.append(pin) 
                    elif asked_camel_types =="Any":
                        corresponding_pins.append(pin)
        return corresponding_pins

    def __str__(self):
        docstr = 'Available outputs:\n'
        for output in self._outputs:
            tot_string = str(output._spec.name)
            input_string = tot_string.split('\n')
            input_string1 = input_string[0]
            line = ["   ", "- ", input_string1]
            docstr += '{:<5}{:<4}{:<20}\n'.format(*line)
            for inputstr in input_string:
                if inputstr != input_string1:
                    line = ["   ", "  ", inputstr]
                    docstr += '{:<5}{:<4}{:<20}'.format(*line)
                    docstr += '\n'
        return docstr
            
def _clearRepeatedMessage(message):
    try:
        while True:
            message.pop(len(message)-1)
    except:
        pass
    
def _make_printable_type(type):
    """Remove character not allowed in function names in python"""
    chars_to_remove = ['<', '>', '-', '!', '?', "."]
    rx = '[' + re.escape('_'.join(chars_to_remove)) + ']'
    type = re.sub(rx, '_', type)
    type = type.replace("abstract_","")
    return type

def _modify_output_spec_with_one_type(output_spec, type):
    if isinstance(output_spec, operator_pb2.PinSpecification):
        spec = operator_pb2.PinSpecification() #create a copy of the pin spec with only one type
        spec.CopyFrom(output_spec)     
        _clearRepeatedMessage(spec.type_names)
        spec.type_names.extend([type])
    else:
        spec = output_spec
        spec.type_names = [type]
    
    return spec

class Outputs(_Outputs):
    def __init__(self, dict_outputs, operator):
        super().__init__(dict_outputs, operator)
        for pin in self._dict_outputs:
            if len(self._dict_outputs[pin].type_names)==1 and self._dict_outputs[pin] is not None:
                output_name = self._dict_outputs[pin].name
                output = Output(self._dict_outputs[pin], pin, self._operator)
                self._outputs.append(output)
                if hasattr(self, output_name):  # for autogenerated code
                    setattr(self, '_' + output_name, output)
                else:
                    setattr(self, output_name, output)
            #generate 1 output by type name
            elif len(self._dict_outputs[pin].type_names)>1 and self._dict_outputs[pin] is not None:
                for type in self._dict_outputs[pin].type_names:                    
                    output_name = self._dict_outputs[pin].name +"_as_"+_make_printable_type(type)                 
                    output = Output(_modify_output_spec_with_one_type(self._dict_outputs[pin], type), pin, self._operator)
                    self._outputs.append(output)
                    if hasattr(self, output_name):  # for autogenerated code
                        setattr(self, '_' + output_name, output)
                    else:
                        setattr(self, output_name, output)



"""
nodal_from_mesh
===============
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "scoping" category
"""

class nodal_from_mesh(Operator):
    """Get the nodes ids scoping of an input mesh.

      available inputs:
        - mesh (MeshedRegion)

      available outputs:
        - mesh_scoping (Scoping)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.scoping.nodal_from_mesh()

      >>> # Make input connections
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.scoping.nodal_from_mesh(mesh=my_mesh)

      >>> # Get output data
      >>> result_mesh_scoping = op.outputs.mesh_scoping()"""
    def __init__(self, mesh=None, config=None, server=None):
        super().__init__(name="GetNodeScopingFromMesh", config = config, server = server)
        self._inputs = InputsNodalFromMesh(self)
        self._outputs = OutputsNodalFromMesh(self)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Get the nodes ids scoping of an input mesh.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "mesh_scoping", type_names=["scoping"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "GetNodeScopingFromMesh")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsNodalFromMesh 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsNodalFromMesh 
        """
        return super().outputs


#internal name: GetNodeScopingFromMesh
#scripting name: nodal_from_mesh
class InputsNodalFromMesh(_Inputs):
    """Intermediate class used to connect user inputs to nodal_from_mesh operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.scoping.nodal_from_mesh()
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(nodal_from_mesh._spec().inputs, op)
        self._mesh = Input(nodal_from_mesh._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.scoping.nodal_from_mesh()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsNodalFromMesh(_Outputs):
    """Intermediate class used to get outputs from nodal_from_mesh operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.scoping.nodal_from_mesh()
      >>> # Connect inputs : op.inputs. ...
      >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """
    def __init__(self, op: Operator):
        super().__init__(nodal_from_mesh._spec().outputs, op)
        self._mesh_scoping = Output(nodal_from_mesh._spec().output_pin(0), 0, op) 
        self._outputs.append(self._mesh_scoping)

    @property
    def mesh_scoping(self):
        """Allows to get mesh_scoping output of the operator


        Returns
        ----------
        my_mesh_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.scoping.nodal_from_mesh()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh_scoping = op.outputs.mesh_scoping() 
        """
        return self._mesh_scoping




from ansys.dpf.core.server import DpfServer
from ansys.dpf import core
from ansys.dpf.core import LOCALHOST
import weakref

"""Aeneid specific functions and classes"""

def start_server_using_service_manager():  # pragma: no cover
    if core.module_exists("grpc_interceptor_headers"):
        import grpc_interceptor_headers
        from  grpc_interceptor_headers.header_manipulator_client_interceptor import header_adder_interceptor    
    else:
        raise ValueError('Module grpc_interceptor_headers is missing to use service manager, please install it using pip install grpc_interceptor_headers')

    service_manager_url = f"http://{LOCALHOST}:8089/v1"

    definition = requests.get(url=service_manager_url + "/definitions/dpf").json()
    rsp = requests.post(url=service_manager_url + "/jobs", json=definition)
    job = rsp.json()

    dpf_task = job['taskGroups'][0]['tasks'][0]
    dpf_service = dpf_task['services'][0]
    dpf_service_name = dpf_service['name']
    dpf_url = f"{dpf_service['host']}:{dpf_service['port']}"

    channel = grpc.insecure_channel(dpf_url)
    header_adder =  header_adder_interceptor('service-name', dpf_service_name)
    intercept_channel = grpc.intercept_channel(channel, header_adder)
    core.SERVER = DpfJob(service_manager_url, dpf_service_name,intercept_channel)

    core._server_instances.append(weakref.ref(core.SERVER))


class DpfJob(DpfServer):
    def __init__(self, service_manager_url, job_name, channel):
        self.sm_url = service_manager_url
        self.job_name = job_name
        super().channel = channel

    def shutdown(self):
        requests.delete(url=f'{self.sm_url}/jobs/{self.job_name}')

    def __del__(self):
        try:
            self.shutdown()
        except:
            pass



# -*- coding: utf-8 -*-
"""
MeshesContainer
===============
Contains classes associated to the DPF MeshesContainer
"""
from ansys import dpf
from ansys.dpf.core.collection import Collection
from ansys.dpf.core.common import types


class MeshesContainer(Collection):
    """A class used to represent a MeshesContainer which contains
    meshes split on a given space

    Parameters
    ----------
    meshes_container : ansys.grpc.dpf.collection_pb2.Collection or ansys.dpf.core.MeshesContainer, optional
        Create a meshes container from a Collection message or create a copy from an existing meshes container

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
    """

    def __init__(self, meshes_container=None, server=None):
        """Initialize the scoping with either optional scoping message,
        or by connecting to a stub.
        """
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()
        
        Collection.__init__(self, types.meshed_region,  
                            collection=meshes_container, server=self._server)

    def get_meshes(self, label_space):
        """Returns the meshes at a label space

        Parameters
        ----------
        label_space : dict[str,int] 
            Meshes corresponding to the filter (label space) in input, for example:
            ``{"elshape":1, "body":12}``

        Returns
        -------
        meshes : list[MeshedRegion]
            meshes corresponding to the request
        """
        return super()._get_entries(label_space)
    
    def get_mesh(self, label_space_or_index):
        """Returns the mesh at a requested index or label space.
        Throws if the request returns several meshes

        Parameters
        ----------
        label_space_or_index : dict[str,int] , int 
            Scoping of the requested mesh, for example:
            ``{"time": 1, "complex": 0}``
            or Index of the mesh.

        Returns
        -------
        mesh : MeshedRegion
            mesh corresponding to the request
        """
        return super()._get_entry(label_space_or_index)

    def __getitem__(self, key):
        """Returns the mesh at a requested index

        Parameters
        ----------
        key : int
            the index

        Returns
        -------
        mesh : MeshedRegion
            mesh corresponding to the request
        """
        return super().__getitem__(key)

    def add_mesh(self, label_space, mesh):
        """Update or add the scoping at a requested label space.

        Parameters
        ----------
        label_space : dict[str,int]
            label_space of the requested meshes, ex : {"elshape":1, "body":12}

        mesh : MeshedRegion
            DPF mesh to add.
        """
        return super()._add_entry(label_space, mesh)

    def __str__(self):
        txt = 'DPF Meshes Container with\n'
        txt += "\t%d mesh(es)\n" % len(self)
        txt += f"\tdefined on labels {self.labels} \n\n"
        return txt

    



"""
PropertyField
=============
"""

from ansys.dpf.core.common import natures, locations
from ansys.dpf.core.field_base import _FieldBase,_LocalFieldBase

class PropertyField(_FieldBase):
    """Property field class is a Field that describes properties such as connectivity. 
    It is a Field with integer values instead of double values.
    
    Parameters
    ----------
    nentities: int
        Number of entities that the property field will contain. 
        
    nature: core.natures
        Nature of the field (scalar, vector...)
        
    Returns
    -------
    property_field: PropertyField
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> pfield = dpf.PropertyField()
    >>> list_ids = [1, 2, 4, 6, 7]
    >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
    >>> pfield.scoping = scop
    >>> list_data = [20, 30, 50, 70, 80]
    >>> pfield.data = list_data
    
    """
    
    def __init__(self, nentities=0, nature=natures.scalar, location=locations.nodal, 
                 property_field=None, server=None):
        super().__init__(nentities, nature, location, True, property_field, server)
        
    @property
    def location(self):
        """Return the property field location. Property field contains a scoping and this 
        is the location that is read. To update location, directly update the scoping 
        location. 

        Returns
        -------
        location : str
            Location string.  Either ``'Nodal'`` or ``'Elemental'``.

        Examples
        --------
        Create a property field and request location
        
        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.PropertyField()
        >>> list_ids = [1, 2, 4, 6, 7]
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.scoping.location = dpf.locations.nodal
        >>> pfield.location
        'Nodal'
        
        """
        if self.scoping:
            return self.scoping.location
        else:
            return None
        
    @location.setter
    def location(self, value):
        """Change the property field location.

        Parameters
        ----------
        location : str or locations
            Location string.  Either ``'Nodal'``, ``'Elemental'``.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> pfield = dpf.PropertyField()
        >>> scop = dpf.Scoping(ids = list_ids, location = dpf.locations.nodal)
        >>> pfield.scoping = scop
        >>> pfield.location = 'Nodal'
        >>> pfield.location
        'Nodal'
        
        """
        if self.scoping:
            self.scoping.location = value
        else:
            raise Exception("Property field location is based on scoping, and scoping is not defined")
            
    def as_local_field(self):
        """Creates a deep copy of the field locally so that the user can access 
        and modify it locally without any request sent to the server.
        This should be used in a with statement, so that the local field
        is released and the data is sent to the server in one shot.
        If it's not used in a with statement, the method release_data()
        should be used to actually update the field.
        
        Warning
        -------
        If this as_local_field method is not used as a context manager in a 
        with statement or if the method release_data() is not called,
        the data will not be actually updated.
        
        Returns
        -------
        local_field : PropertyField
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> num_entities = 5
        >>> field_to_local = dpf.PropertyField(num_entities, dpf.natures.scalar)
        >>> with field_to_local.as_local_field() as f:    
        ...     for i in range(1,num_entities+1):
        ...         f.append(list(range(i,i+3)),i)
        ...         f.get_entity_data(i-1)
        array([1, 2, 3])
        array([2, 3, 4])
        array([3, 4, 5])
        array([4, 5, 6])
        array([5, 6, 7])
        
        """
        return _LocalPropertyField(self)  


class _LocalPropertyField(_LocalFieldBase,PropertyField):
    """Class only created by a field to cache the internal data of the field,
    modify it locallly, and send a single update request to the server 
    when the local field is deleted
    
    Parameters
    ----------
    field : PropertyField
        property field to copy

    """
    
    def __init__(self, field):
        super().__init__(field)
        


"""
mesh_scoping_factory
====================

Contains functions to make easy mesh scopings creation.
"""

from ansys.dpf import core
from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import Scoping
from ansys.dpf.core import errors as dpf_errors

def nodal_scoping(node_ids, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``
    associated to a mesh. 

    Parameters
    ----------
    node_ids : List of int
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
    """
    if not isinstance(node_ids, list):
        raise dpf_errors.InvalidTypeError("list", "node_ids") 
    scoping = Scoping(server = server, ids = node_ids, location = locations.nodal)
    return scoping

def elemental_scoping(element_ids, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``
    associated to a mesh. 

    Parameters
    ----------
    element_ids : List of int
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
    """
    if not isinstance(element_ids, list):
        raise dpf_errors.InvalidTypeError("list", "element_ids") 
    scoping = Scoping(server = server, ids = element_ids, location = locations.elemental)
    return scoping

def named_selection_scoping(named_selection_name, model, server = None):
    """Helper function to create a specific ``ansys.dpf.core.Scoping``
    associated to a specified model's mesh. 

    Parameters
    ----------
    named_selection_name : str
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    scoping : ansys.dpf.core.Scoping
    """
    return model.metadata.named_selection(named_selection_name)